# import sys
# sys.path.append("..")

from profq_data.data.binary_tree import BinaryTree
from profq_data.data.doubly_linked_list import DoublyLinkedList
from profq_data.data.singly_linked_list import SinglyLinkedList
from profq_data.data.stack import Stack
from profq_data.data.queue import Queue
from profq_data.data.priority_queue import PriorityQueue
from profq_data.data.skiplist import SkipList
from profq_data.data.hashtable import HashTable
from profq_data.data.heap import Heap


class LinkedList:
    """A base class for singly and doubly linked lists
    """
    def __init__(self) -> None:
        """The init function
        """
        self.head = None

        self._size = 0
        
    def peek(self) -> int:
        """Peek the head value of the linked lsit

        Returns:
            int: the value of the head
        """
        return self.head.data
    
    def __len__(self) -> int:
        """Gets the size of the linked list

        Returns:
            int: the size of the linked list
        """
        return self._size
    
    def contains(self, data: int) -> bool:
        """Tests if the linked list contains data

        Args:
            data (int): the data to find in the list

        Returns:
            bool: if the data is contained in the list
        """
        # If the list is empty
        if self.head is None:
            return

        # If data from head is equal to data, return True
        if self.head.data == data:
            return True

        # Loop over all the items in the linked list
        current = self.head
        while current is not None:
            # Return true if the current's data is equal to the to search data
            if current.data == data:
                return True
            
            current = current.next
        
        # Return false if it doesn't contain
        return False
    
    
    def print_items(self):
        """Print the list
        """
        if self.head is None:
            return

        current = self.head
        while current is not None:
            print(current.data, end=" -> ")

            current = current.next
        
        print()


from profq_data.helpers.linked_list import LinkedList
from profq_data.helpers.nodes.skiplist_node import Node

class SkipList_LinkedList(LinkedList):
    """An implementation of a singly linked list
    """
    def push(self, data: int):
        """Prepend the data to the start of the array

        Args:
            data (int): the data to prepend
        """
        new_head = Node(data)

        # Point the new head to the previous head
        new_head.next = self.head
        
        # Set the new head
        self.head = new_head

        self._size += 1

    def append(self, data: int):
        """Append data to the end of the array

        Args:
            data (int): the data to append
        """
        if self.head is None:
            self.head = Node(data)
            return

        # Loop over the array until the end
        current = self.head
        while current.next is not None:
            current = current.next

        # Add the new node
        current.next = Node(data)
        
        self._size += 1
    
    def delete_with_value(self, data: int):
        """Delete a node with a certain value

        Args:
            data (int): the value of the node to be deleted
        """
        # If the list is empty
        if self.head is None:
            return
            
        # If data from head is equal to data, remove data
        if self.head.data == data:
            self.head = self.head.next
            self._size -= 1
            return
        
        # Loop over the list to find the item with the data
        current = self.head
        while current.next is not None:
            # If next data is equal to data
            # Reroute the next to the next next
            # "Skip" the next node
            if current.next.data == data:
                current.next = current.next.next
                break
            
            current = current.next
            
        self._size -= 1

    def delete_at_index(self, index: int) -> int:
        """Delete an item at index

        Args:
            index (int): the index for the item to delete

        Raises:
            IndexError: when the index is not in the range of the linked list

        Returns:
            int: the value of the node at index
        """
        # If the list is empty
        if self.head is None:
            return
        
        # Test if index is in range of the linked list
        if index < 0 or index > self._size - 1:
            raise IndexError(f"Index {index} not in linked list {self}")

        # If data from head is equal to data, remove data
        if index == 0:
            self.head = self.head.next
            self._size -= 1
            return

        # Loop over the list to find the item with the data
        current = self.head
        for _ in range(index):
            current = current.next
        
        # Reroute the next to the next next
        # "Skip" the next node
        current.next = current.next.next
        
        self._size -= 1



# sourcery skip: path-read
from setuptools import setup, find_packages

VERSION = '1.0.0'
DESCRIPTION = 'A library for data structures.'


with open("README.md") as f:
    LONG_DESCRIPTION = f.read()

# Setting up
setup(
    name="profq_data",
    version=VERSION,
    author="ProfessorQu",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    keywords=['python', 'data structures', 'algorithms'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Microsoft :: Windows",
    ]
)


class HeapCalculator:
    """A helper class for the heap to calculate indeces
    """
    def __init__(self, heap) -> None:
        """The init function
        """
        self.heap = heap

    def get_left_child_index(self, parent_index: int) -> int:
        """Get the left child index of a parent index

        Args:
            parent_index (int): the parent index

        Returns:
            int: the index of the left child
        """
        return 2 * parent_index + 1
    
    def get_right_child_index(self, parent_index: int) -> int:
        """Get the right child index of a parent index

        Args:
            parent_index (int): the parent index

        Returns:
            int: the index of the right child
        """
        return 2 * parent_index + 2
    
    def get_parent_index(self, child_index: int) -> int:
        """Get the parent index of a child index

        Args:
            child_index (int): the child index

        Returns:
            int: the index of the parent
        """
        return (child_index - 1) // 2
    
    def has_left_child(self, parent_index: int) -> bool:
        """Check if a parent index has a left child

        Args:
            parent_index (int): the parent index

        Returns:
            bool: whether the parent has a left child
        """
        return self.get_left_child_index(parent_index) < len(self.heap)
    
    def has_right_child(self, parent_index: int) -> bool:
        """Check if a parent index has a right child

        Args:
            parent_index (int): the parent index

        Returns:
            bool: whether the parent has a right child
        """
        return self.get_right_child_index(parent_index) < len(self.heap)
    
    def has_parent(self, child_index: int) -> bool:
        """Check if a child index has a parent

        Args:
            child_index (int): the child index

        Returns:
            bool: whether the child has a parent
        """
        return self.get_parent_index(child_index) >= 0
    
    def get_left_child(self, parent_index: int) -> int:
        """Get the left child of a parent index

        Args:
            parent_index (int): the parent index

        Returns:
            int: the left child
        """
        return self.heap[self.get_left_child_index(parent_index)]
    
    def get_right_child(self, parent_index: int) -> int:
        """Get the right child of a parent index

        Args:
            parent_index (int): the parent index

        Returns:
            int: the right child
        """
        return self.heap[self.get_right_child_index(parent_index)]
    
    def get_parent(self, child_index: int) -> int:
        """Get the parent of a child index

        Args:
            child_index (int): the child index

        Returns:
            int: the parent
        """
        return self.heap[self.get_parent_index(child_index)]
    

